from . import comm
import matplotlib.pyplot as plt
from typing import List
from matplotlib.ticker import ScalarFormatter
from matplotlib.ticker import MaxNLocator

#try:
#    from matplotlib import rc
#    rc('font',**{'family':'sans-serif'})
#    rc('text', usetex=True)
#except:
#    print("No latex support in matplotlib.")

def check_font_exists(font_name):
    import matplotlib.font_manager as fm
    available_fonts = [f.name for f in fm.fontManager.ttflist]
    return font_name in available_fonts


def set_font(font_name:str="Times New Roman", 
             font_size:int=None):
    """Set the font for matplotlib.

    Args:
        font_name (str, optional): font name. Defaults to "Times New Roman".
        font_size (int, optional): font size. Defaults to 18.
    """
    try:
        if not check_font_exists(font_name):
            print(f"Font '{font_name}' not found. Using default font.")
            return

        plt.rcParams["font.family"] = font_name  
        
        plt.rcParams["mathtext.fontset"] = "custom"
        plt.rcParams["mathtext.rm"] = font_name
        if check_font_exists(f"{font_name}:italic"):
            plt.rcParams["mathtext.it"] = f"{font_name}:italic"
        if check_font_exists(f"{font_name}:bold"):
            plt.rcParams["mathtext.bf"] = f"{font_name}:bold"
    except Exception as e:
        pass

    if font_size is not None:
        plt.rcParams["font.size"] = font_size


def plot_line_point(x,
                    ys,
                    legends:List[str]=None, 
                    title: str=None, 
                    xtitle: str=None, 
                    ytitle: str=None,
                    xlabels:List[str]=None,
                    fname:str=None,
                    figsize:tuple=(8,6),
                    fontsize:int=18,
                    colors:List[str] =None,
                    markers:List[str] =None,
                    grid:bool=True,
                    legend_outbox:bool=False):
    """Plot x vs ys[i] with line and point.

    Args:
        x (List[str|float|int]): x values.
        ys (List[List[float|int]]): y values.
        legends (List[str], optional): legends. Defaults to None.
        title (str, optional): title. Defaults to None.
        xtitle (str, optional): x title. Defaults to None.
        ytitle (str, optional): y title. Defaults to None.
        xlabels (List[str], optional): x labels. Defaults to None.
        fname (str, optional): file name. Defaults to None.
        figsize (tuple, optional): figure size. Defaults to (8,6).
        fontsize (int, optional): font size. Defaults to 18.
        colors (List[str], optional): colors. Defaults to None.
        markers (List[str], optional): markers. Defaults to None.
        grid (bool, optional): grid. Defaults to True.
        legend_outbox (bool, optional): legend outbox. Defaults to False
    """
    if colors == None:
        colors = ["b","g","r","c","m","y","k"]
    if markers == None:
        markers = ["o","*","+","x","s","p","d"]

    # if x is a list of string, then convert it to a list of int, and set xlabels to x
    if isinstance(x[0],str):
        if xlabels is None:
            xlabels = x
        x = list(range(len(x)))
    
    plt.figure(figsize=figsize)
    for i in range(len(ys)):
        if len(x) != len(ys[i]):
            print("Lengths of x and ys[{}] are different.".format(i))
            continue    

        ix,iy = comm.clean_none_list(x,ys[i])
        if len(ix) == 0:
            print("All elements in x and ys[{}] are None.".format(i))
            continue
        
        if legends is not None:
            ilegend = legends[i]
        else:
            ilegend = None
        
        plt.plot(ix,iy, label=ilegend, color=colors[i%len(colors)], marker=markers[(i//len(colors))%len(markers)])

    if title is not None:
        plt.title(title, fontsize=fontsize)
    if xtitle is not None:
        plt.xlabel(xtitle, fontsize=fontsize)
    if ytitle is not None:
        plt.ylabel(ytitle, fontsize=fontsize)
    if legends is not None:
        if legend_outbox:
            plt.legend(fontsize=fontsize-2, loc="upper left", bbox_to_anchor=(1, 1))
        else:
            plt.legend(fontsize=fontsize-2)
    if xlabels is not None:
        plt.xticks(x, xlabels, fontsize=fontsize-2,rotation=30)
    if grid:
        plt.grid()
    plt.tick_params(axis='x', labelsize=fontsize-2)
    plt.tick_params(axis='y', labelsize=fontsize-2) 
    plt.tight_layout()
    if fname is not None:
        plt.savefig(fname)  
    else:
        plt.show()
    plt.close()
    
def plot_bar(x,
             ys,
             legends:List[str]=None, 
             title: str=None, 
             xtitle: str=None, 
             ytitle: str=None,
             xlabels:List[str]=None,
             fname:str=None,
             figsize:tuple=(8,6),
             fontsize:int=18,
             colors:List[str] =None,
             ref_idx:int=None,
             grid:bool=True,
             legend_outbox:bool=False):
    """Plot x vs ys[i] with bar.

    Args:
        x (List[str|float|int]): x values.
        ys (List[List[float|int]]): y values.
        legends (List[str], optional): legends. Defaults to None.
        title (str, optional): title. Defaults to None.
        xtitle (str, optional): x title. Defaults to None.
        ytitle (str, optional): y title. Defaults to None.
        xlabels (List[str], optional): x labels. Defaults to None.
        fname (str, optional): file name. Defaults to None.
        figsize (tuple, optional): figure size. Defaults to (8,6).
        fontsize (int, optional): font size. Defaults to 18.
        colors (List[str], optional): colors. Defaults to None.
        ref_idx (int, optional): reference index. Defaults to None. If set, then will plot the deviation of ys[i] from ys[ref_idx].
        grid (bool, optional): grid. Defaults to True.
        legend_outbox (bool, optional): legend outbox. Defaults to False
    """
    if colors == None:
        colors = ["b","g","r","c","m","y","k"]
    
    markers = ["o","*","+","x","s","p","d"]

    plt.figure(figsize=figsize)
    for i in range(len(ys)):
        if len(x) != len(ys[i]):
            print("Lengths of x and ys[{}] are different.".format(i))
            continue    

        ix,iy = comm.clean_none_list(x,ys[i])
        if len(ix) == 0:
            print("All elements in x and ys[{}] are None.".format(i))
            continue
        
        if legends is not None:
            ilegend = legends[i]
        else:
            ilegend = None
        
        plt.bar(ix,iy, label=ilegend, color=colors[i%len(colors)])

    if title is not None:
        plt.title(title, fontsize=fontsize)
    if xtitle is not None:
        plt.xlabel(xtitle, fontsize=fontsize)


def plot_boxplot(x, 
                 ys,
                 title: str=None, 
                 xtitle: str=None, 
                 ytitle: str=None,
                 fname:str=None,
                 figsize:tuple=(8,6),
                 fontsize:int=18,):
    """Plot boxplot.

    Args:
        x (List[str|float|int]): x values.
        ys (List[List[float|int]]): y values, the length of ys should be the same as x.
        title (str, optional): title. Defaults to None.
        xtitle (str, optional): x title. Defaults to None.
        ytitle (str, optional): y title. Defaults to None.
        fname (str, optional): file name. Defaults to None.
        figsize (tuple, optional): figure size. Defaults to (8,6).
        fontsize (int, optional): font size. Defaults to 18.
    """
    plt.figure(figsize=figsize)
    plt.boxplot(ys, labels=x)
    if title is not None:
        plt.title(title, fontsize=fontsize)
    if xtitle is not None:
        plt.xlabel(xtitle, fontsize=fontsize)
    if ytitle is not None:
        plt.ylabel(ytitle, fontsize=fontsize)
    
    if fname is not None:
        plt.savefig(fname)
    else:
        plt.show()
    plt.close()


def gen_force_fd_1(ax1, ax2, x, e, ana, fd,
                   x_title = "Position ($\mathrm{\AA}$)",
                   e_title = "Energy (eV)",
                   f_title = "Force (eV/$\mathrm{\AA}$)",
                   e_label = "Energy (eV)",
                   ana_label = "Analytic (eV/$\mathrm{\AA}$)",
                   fd_label = "Finite Difference (eV/$\mathrm{\AA}$)",
                   title = None,
                   font_size = None,
                   grid = True
                   ):
    """
    Generate the force vs distance plot for the first case.
    
    """
    d,e,f,fd = comm.clean_none_list(x, e, ana, fd)
    
    if len(d) == 0 or len(x) == 0:
        return None

    if font_size is None:
        font_size = 18
    fz_label = font_size
    fz_xy_title = font_size + 4
    fz_title = None if font_size is None else font_size + 4
    fz_tick = None if font_size is None else font_size
    fz_marker = None if font_size is None else font_size - 6

    ax1.plot(d,e,"b+-",label=e_label,markersize=fz_marker, linewidth=3)
    ax2.plot(d,f,"ro-",label=ana_label,markersize=fz_marker, linewidth=3)
    ax2.plot(d,fd,"m*--",label=fd_label,markersize=fz_marker, linewidth=3)
    
    ymax1 = max(e)
    ymin1 = min(e)
    ymax2 = max(max(f),max(fd))
    ymin2 = min(min(f),min(fd))
    ax1.set_xlabel(x_title,fontsize=fz_xy_title)
    ax1.set_ylabel(e_title,color="b",fontsize=fz_xy_title)
    ax2.set_ylabel(f_title,color="r",fontsize=fz_xy_title)
    ax1.spines['left'].set_color('blue')
    ax1.tick_params(axis='y', colors='blue')
    ax2.spines['right'].set_color('red')
    ax2.tick_params(axis='y', colors='red')
    ax1.set_ylim(ymin1-(ymax1-ymin1)*0.1,ymax1+(ymax1-ymin1)*0.2)
    ax2.set_ylim(ymin2-(ymax2-ymin2)*0.1,ymax2+(ymax2-ymin2)*0.2)
    ax1.legend(loc="upper left",fontsize=fz_label)
    ax2.legend(loc="upper right",fontsize=fz_label)
    if title is not None:
        ax1.set_title(title,fontsize=fz_title)
    if grid:
        ax2.grid(True)
    if fz_tick is not None:
        ax1.tick_params(axis='x', labelsize=fz_tick, direction='in')
        ax1.tick_params(axis='y', labelsize=fz_tick, direction='in')
        ax2.tick_params(axis='y', labelsize=fz_tick, direction='in')
        
    ax1.yaxis.set_major_locator(MaxNLocator(nbins=5))
    
    ax1.yaxis.get_offset_text().set_visible(False)
    ax1.yaxis.set_major_formatter(ScalarFormatter(useOffset=False))
    ax2.yaxis.get_offset_text().set_visible(False)
    ax2.yaxis.set_major_formatter(ScalarFormatter(useOffset=False))
    

def auto_set_yaxis(ax, ys, margin1=0.1, margin2=0.3, log_scale_threshold=1000):
    """Auto set y limit.

    Args:
        ax (matplotlib.axes._subplots.AxesSubplot): axes.
        ys (List[float]): y values.
        margin (float, optional): margin. Defaults to 0.1.
    """
    ymax = max(ys)
    ymin = min(ys)
    
    ax.yaxis.get_offset_text().set_visible(False)
    ax.yaxis.set_major_formatter(ScalarFormatter(useOffset=False))
    # if the y values has a large range, then use scientific notation, and set log scale
    if ymax * ymin > 0 and ymax/ymin > log_scale_threshold:
        ax.set_yscale('log')
        ymax = ymax * 10
        ax.set_ylim(ymin,ymax)
        #ax.yaxis.set_major_formatter(ScalarFormatter(useOffset=False))
        #ax.yaxis.set_minor_formatter(ScalarFormatter(useOffset=False))
        #ax.yaxis.set_major_locator(MaxNLocator(nbins=5))
    else:
        ymin = ymin - (ymax-ymin)*margin1
        ymax = ymax + (ymax-ymin)*margin2
        ax.set_ylim(ymin,ymax)
        ax.yaxis.set_major_locator(MaxNLocator(nbins=5))
    return ymin, ymax
    
    


import os, json, traceback, copy

def csv2table(csvfile):
    '''
    Transform a csv file to a table
    '''
    if not os.path.exists(csvfile):
        print(f"Error: {csvfile} does not exist!")
        return []
    with open(csvfile) as f:
        lines = f.readlines()
        table = []
        for line in lines:
            table.append(line.strip().split(","))
    return table

def json2table(jsonfile):
    '''
    Transform a json file to a table
    The key should be the first column that is the example name, and the value should be a dict, which are the other columns, and the keys are the metric name.
    
    '''
    if not os.path.exists(jsonfile):
        print(f"Error: {jsonfile} does not exist!")
        return []
    values = json.load(open(jsonfile))
    table = []
    metrics = []
    for k,v in values.items():
        for ik in v.keys():
            if ik not in metrics:
                metrics.append(ik)

    table.append(["example"] + metrics)
    for ikey in values.keys():
        table.append([ikey] + [values[ikey].get(imetric,None) for imetric in metrics])
    
    return table

def json2table_sm(jsonfile):
    '''
    Transform a json file to a table
    The key should be the first column that is supermetrics name, and the value is the value.
    '''
    if not os.path.exists(jsonfile):
        print(f"Error: {jsonfile} does not exist!")
        return []
    values = json.load(open(jsonfile))
    table = []
    for k,v in values.items():
        if isinstance(v,(int,float,str,bool,type(None))):
            table.append([k,v])
        else:
            print(f"Warning: type of the value of {k} is {type(v)}, which is not supported. Ignored.")
    
    return table

def file2table(metricfile):
    # based on the file type to read the file
    if not os.path.exists(metricfile):
        print(f"Error: {metricfile} does not exist!")
        return None
    
    filetype = os.path.splitext(metricfile)[1]
    if filetype == ".csv":
        try:
            table = csv2table(metricfile)
        except:
            traceback.print_exc()
            print(f"Error: transfer {metricfile} to table failed!")
            return None
    elif filetype == ".json":
        try:
            table = json2table(metricfile)
        except:
            traceback.print_exc()
            print(f"Error: transfer {metricfile} to table failed!")
            return None
    else:
        print(f"Error: file '{filetype}' of table is not supported!")
        return None
    return table

def rotate_table(table):
    '''
    Rotate a table
    '''
    new_table = []
    for i in range(len(table[0])):
        new_table.append([table[j][i] for j in range(len(table))])
    return new_table

def isort(itable_input,head_list):
    itable = copy.deepcopy(itable_input)
    heads = copy.deepcopy(itable[0])
    sort_idx = []
    for i in head_list:
        if i not in heads:
            print("isort() ERROR:",i,"is not a head of input table. Table head is:",heads)
            return itable
        else:
            sort_idx.append(heads.index(i))
    jtable = []
    for i in itable[1:]:
        jtable.append([i[j] for j in sort_idx] + [i])
    jtable.sort()
    return [heads] + [i[-1] for i in jtable]

def output_float(f, prec=2):
    '''
    Output a string of float with precision, default 2
    If f is a small number, output in scientific notation
    '''
    if f == None:
        return "---"
    elif isinstance(f, str):
        return f
    elif isinstance(f, int):
        return str(f)
    
    try:
        f = float(f)
    except:
        return f
    
    if abs(f) < pow(10,-1*prec):
        return '%.2e' % f
    else:
        return '%.*f' % (prec, f)
    
def judge_metric(x, criteria):
    '''
    Judge if a metric is good or bad based on criteria
    '''
    try:
        x = float(x)
        sm_pass = eval(criteria)
        return bool(sm_pass)
    except:
        return None
    
def format_table(table,metrics_name=None, sort=None, criteria=None, color={True:"green",False:"red"}):
    '''
    table: a list of list, each list is a row of the table. The first row is the head of the table
    metrics_name: a list of metrics name, which will be output in the table
    sort: a list of metrics name, which will be used to sort the table
    criteria: a dict of criteria, the key is the metric name, and the value is the criteria
    
    Do things:
    1. If the value is a number, then round it to 4 digits
    2. If the value pass the criteria, the value will be colored to green, else red.
    3. If the value is None, then transfer to "---"
    
    criteria is a dict:
    {
        "key": "key_name",
        "criteria": "x > 0"   # this is a string, should be evaluated by python
    }
    
    return a table, and a dict of pass number
    {
        "key_metrics": {"pass": 1, "notpass": 2},
        "all": {"pass": 3, "total": 4}
    }
    '''
    print("criteria:",criteria)
    print("table head:",table[0])
    
    if metrics_name in [[],None]:
        metrics_name = table[0]
    if table[0][0] not in metrics_name:
        metrics_name = [table[0][0]] + metrics_name # the first colume should be the example name
    
    metric_idx = []
    for i in metrics_name:
        if i not in table[0]:
            metric_idx.append(None)
        else:
            metric_idx.append(table[0].index(i))
    
    new_table = [metrics_name]
    for i in range(1,len(table)):
        new_table.append([None if j == None else table[i][j] for j in metric_idx])
    
    if sort:
        new_table = isort(new_table,sort)    
    table = new_table
        
    pass_num = {k:{"pass":0,"notpass":0} for k in criteria.keys()}
    pass_num["all"] = {"pass":0,"total":len(table)-1}
    for i in range(1,len(table)):
        allpass = True
        for j in range(len(table[i])):
            metric_name = table[0][j]
            if metric_name in criteria:
                metric_pass = judge_metric(table[i][j], criteria[metric_name])
                if metric_pass == True:
                    table[i][j] = '<font color="%s">%s</font>' % (color[True], output_float(table[i][j]))
                    pass_num[metric_name]["pass"] += 1
                else:
                    table[i][j] = '<font color="%s">%s</font>' % (color[False], output_float(table[i][j]))
                    if metric_pass == False:
                        pass_num[metric_name]["notpass"] += 1
                    allpass = False
            else:
                table[i][j] = output_float(table[i][j])
        if allpass:
            pass_num["all"]["pass"] += 1
                    
    return table, pass_num

def gen_criteria(criteria,pass_num):
    html =  f'''
    <table border="2px">
        <tbody>
            <tr>
                <td>Key</td>
                <td>Pass/Total</td>
                <td>Criteria</td>
            </tr>
        '''
    for key in criteria.keys():
        if key not in pass_num:
            continue
        icolor = "green" if pass_num[key]["pass"] == pass_num[key]["pass"]+pass_num[key]["notpass"] else "red"
        html += f'''<tr>
                <td>{key}</td>
                <td style="color:{icolor}">{pass_num[key]["pass"]}/{pass_num[key]["pass"]+pass_num[key]["notpass"]}</td> 
                <td>{criteria[key]}</td>
            </tr>
        '''
        
    html += '''</tbody>
    </table>\n\n'''
    return html

def gen_criteria_sm(criteria,sm):
    '''
    sm = {key: value}
    '''
    html =  f'''
    <table border="2px">
        <tbody>
            <tr>
                <td>Metric</td>
                <td>Value</td>
                <td>Criteria</td>
            </tr>
        '''
    for ik,iv in sm.items():
        if ik in criteria:
            icolor = "green" if judge_metric(iv, criteria[ik]) else "red"
            value = output_float(iv)
            html += f'''<tr>
                    <td>{ik}</td>
                    <td style="color:{icolor}">{value}</td> 
                    <td>{criteria[ik]}</td>
                </tr>
            '''
        else:
            html += f'''<tr>
                    <td>{ik}</td>
                    <td>{output_float(iv)}</td> 
                    <td>---</td>
                </tr>
            '''
        
    html += '''</tbody>
    </table>\n\n'''
    return html


from abc import ABC, abstractmethod
from pymatgen.io.vasp.inputs import Incar, Poscar
from pymatgen.io.vasp.outputs import Oszicar, Outcar, Vasprun
from pymatgen.core.periodic_table import Element
from pathlib import Path
import pandas as pd
from typing import List, Dict, Literal, Union

class BasicProperty(ABC):
    def __init__(self, magnetization: tuple = None, 
                 energy:float = None, 
                 volume: float=None, 
                 structrue: Poscar = None):
        self._magnetization = magnetization
        self.energy = energy
        self.volume = volume
        self.structure = structrue

    @abstractmethod
    def bond_length(
            self, 
        ) -> pd.DataFrame:
        r"""Calculated bond length of all transition metal atoms to their nearest neighbors 

        Returns
        -------
        - dataframe : (`pandas.DataFrame`) The bond length of all transition metal atoms 
                to their nearest neighbors
        """
        pass

    @abstractmethod
    def magnetic_moment(
            self, 
            element: Literal['TM', 'O', 'ALL']='ALL'
        ) -> pd.DataFrame:
        r"""Grasp the magnetic moment of elements interested
        Parameters
        ----------
        - element : (`str`) interested element, now support 'TM', 'O'(oxygen) and 'ALL', stands for exporting
                magnetic moment of 'Transition Metal', 'Oxygen' and 'ALL element'

        Returns
        -------
        - mag : (`pandas.DataFrame`) magnetization of interested element
        """
        pass

    @abstractmethod
    def inter_layer_distance(
        self,
        ) -> dict:
        r"""Calculate inter layer distance of both TM-layer and alkali-layer in layered structure

        Returns
        -------
        - argument : (`Dict[float]`) Dictionary of averaged inter layer distance
                    e.g. {"TM-layer": 2.5, "Alkali-layer": 2.1}
        """
        pass

    def __cut(self, sortedList: List[float], num: float) -> List[List[float]]:
        r"""Cut a sorted list into List[List] by the distance of its elements
            e.g. [1,1,1,3,3,7] -> [[1,1,1], [3,3], [7]] if num <= 1
            e.g. [1,1,1,3,3,7] -> [[1,1,1,3,3],[7]] if 2 < num <= 4

        Parameters
        ----------
        - sortedList : (`List`) A sorted list
        - num : (`List`) A float decide how the sorted list devided

        Returns
        -------
        - argument : (`List[List[float]]`)
        """
        ptr = 0
        list = []
        for i in range(0, len(sortedList) - 1):
            if (sortedList[i+1] - sortedList[i]) > num:
                list.append(sortedList[ptr:i+1])
                ptr = i + 1
        list.append(sortedList[ptr::])
        return list
    
    def __get_num_index(self) -> dict:
        r"""Get the number of atoms, index in structure file of all elements in structure

        Returns
        -------
        -dict : (`dict`) Contains element symbol, start index and number of atoms in structure file
                e.g. {"Fe": {"start_index": 0, "number": 20}, 
                      "O": {"start_index": 20, "number":10}, 
                      "F": {"start_index": 30, "number": 10}}
        """
        num_index = {}
        site_symbols = self.structure.site_symbols
        natoms = self.structure.natoms
        index_cnt = 1
        for _id, ele in enumerate(site_symbols):
            num_index[ele] = {"start_index": index_cnt - 1, "number": natoms[_id]}
            index_cnt += natoms[_id]
        return num_index

    def __is_transition(self, ele: str) -> bool:
        TM = ('Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pb', 'Ag', 'Cd')
        return (ele in TM)

    def bond_length(
            self 
        ) -> pd.DataFrame:
        r"""Calculated bond length of all transition metal atoms to their nearest neighbors 

        Returns
        -------
        - dataframe : (`pandas.DataFrame`) The bond length of all transition metal atoms 
                to their nearest neighbors
        """
        structure = self.structure.structure
        num_index = self.__get_num_index()
        TM_bonds = []
        for ele in num_index.keys():
            if self.__is_transition(ele):
                start_index = num_index[ele]["start_index"]
                number = num_index[ele]["number"]
                TM_sites = structure.sites[start_index:start_index + number]
                neighbor = structure.get_all_neighbors(r=2.5, sites=TM_sites)
                for ii in range(len(TM_sites)):
                    temp_dict = {"Total_index": start_index + ii + 1, "Atomic_index": ii + 1, "Element": ele}
                    total_length = 0
                    for jj in range(len(neighbor[ii])):
                        length = TM_sites[ii].distance(neighbor[ii][jj])
                        temp_dict["bond_length_{}".format(jj + 1)] = length
                        total_length += length
                    temp_dict["bond_length_ave"] = total_length/(len(neighbor[ii]))
                    TM_bonds.append(temp_dict)
        #return pd.DataFrame.from_dict(TM_bonds)
        return TM_bonds
          
    def magnetic_moment(
            self, 
            element: Literal['TM', 'O', 'ALL']='ALL'
        ) -> list:
        r"""Grasp the magnetic moment of elements interested
        Parameters
        ----------
        - element : (`str`) interested element, now support 'TM', 'O'(oxygen) and 'ALL', stands for exporting
                magnetic moment of 'Transition Metal', 'Oxygen' and 'ALL element'

        Returns
        -------
        - mag : (`pandas.DataFrame`) magnetization of interested element
        """
        if element.upper() not in ('TM', 'O', 'ALL'):
            raise ValueError("Invalid element")
        magnetization = self._magnetization
        num_index = self.__get_num_index()
        for ele in num_index.keys():
            ele_num = num_index[ele].get("number")
            for num in range(ele_num):
                real_index = num_index[ele]["start_index"] + num
                magnetization[real_index]["Total_index"] = real_index + 1
                magnetization[real_index]["Element"] = ele
                magnetization[real_index]["Atomic_index"] = num + 1
        if element == 'ALL':
            return list(magnetization)
            #return pd.DataFrame.from_dict(list(magnetization))
        elif element == 'TM':
            mag = ()
            for ele in num_index.keys():
                if self.__is_transition(ele):
                    start_index = num_index[ele].get("start_index")
                    number = num_index[ele].get("number")
                    mag += magnetization[start_index:start_index + number]
            #return pd.DataFrame.from_dict(list(mag))
            return list(mag)
        else:
            start_index = num_index['O'].get("start_index")
            number = num_index['O'].get("number")
            mag = magnetization[start_index:start_index + number]
            #return pd.DataFrame.from_dict(list(mag))
            return list(mag)

    def inter_layer_distance(self) -> dict:
        from statistics import mean
        r"""Calculated inter-layer distance of layered structure like LiCoO2
        Parameters
        ----------

        Returns
        -------
        - dict : (`dict`) Dictionary contains inter-layer distance of TM-layers and alkali-layers
        """
        cart_coord = self.structure.structure.cart_coords
        num_index = self.__get_num_index()
        start_index = num_index['O'].get("start_index")
        number = num_index['O'].get("number")
        oxy_coord = cart_coord[start_index:start_index+number]
        oxy_coord = sorted(oxy_coord[:,-1])
        cluster_oxy_coord = self.__cut(oxy_coord, 1)
        height_position = []
        for ii in cluster_oxy_coord:
            height_position.append(mean(ii))
        distance = []
        for i in range(1, len(height_position)):
            distance.append(height_position[i] - height_position[i - 1])
        return {"d_TM": min(distance[0], distance[1]), "d_alkali": max(distance[0], distance[1])}

class BasicPropertyVasp(BasicProperty):
    def __init__(self, out_file, struc_file):  
        from pymatgen.io.vasp.outputs import Outcar
        from pymatgen.io.vasp.inputs import Poscar
        self.energy = Outcar(out_file).final_energy
        self.volume = Poscar.from_file(struc_file).structure.volume
        self._magnetization = Outcar(out_file).magnetization
        self.structure = Poscar.from_file(struc_file)
    
if __name__ == "__main__":
    output = BasicPropertyVasp('OUTCAR', 'CONTCAR')
    energy = output.energy
    volume = output.volume
    mag_moment = output.magnetic_moment('TM')
    bond_length = output.bond_length()
    print(energy)
    print(volume)
    print(mag_moment)
    print(bond_length)



import traceback,json
from dp.launching.typing.basic import BaseModel,String
from dp.launching.typing import Field
import os

from abacustest.report import gen_html


from . import (comm_class,
               comm_func,
               comm_report,
               comm_class_exampleSource,
               readsetting)

class CommandSet(BaseModel):
    command: String = Field(default="",
                            title="Command(optional)",
                                    description="bash command if needed",)


class ReportModel(
    #CommandSet,
    #comm_class_exampleSource.ExampleSet,
    comm_class_exampleSource.ExampleSourceSet,
    #comm_class_exampleSource.DatasetSet,
                    comm_class.OutputSet,
                    comm_class.ConfigSet,
                    BaseModel):
    ...  

def ReportModelRunner(opts:ReportModel) -> int:
    try:
        logs = comm_class.myLog()

        paths = comm_func.create_path(str(opts.IO_output_path))
        output_path = paths["output_path"]
        work_path = paths["work_path"]
        download_path = paths["download_path"]

        datas = comm_class_exampleSource.read_source(opts,work_path,download_path,logs.iprint)
        #if opts.command and opts.command.strip() != "":
        #    cwd = os.getcwd()
        #    os.chdir(work_path)
        #    try:
        #        return_code, stdout, stderr = comm_func.run_command(opts.command.strip())
        #        logs.iprint(f"{stdout}\n{stderr}\nrun abacustest over!\n")
        #    except:
        #        traceback.print_exc()
        #        return 1
        #    os.chdir(cwd)
        
        if os.path.isfile(os.path.join(work_path,"setting.json")):
            try:
                setting = json.load(open(os.path.join(work_path,"setting.json")))
            except:
                setting = {}
            if "report" in setting:
                report_setting = setting["report"]  
                pwd = os.getcwd()
                os.chdir(work_path)
                gen_html(report_setting,"abacustest.html")
                os.chdir(pwd)
            
        allparams = {"save_path": "",}


        comm_report.gen_report(opts,logs,work_path,output_path,allparams)
        
    except:
        traceback.print_exc()
        return 1

    return 0



import traceback,json,sys,os
from dp.launching.typing.basic import BaseModel,String,Float,Int,Boolean,Set
from dp.launching.typing import Field
from enum import Enum
from abacustest.lib_model.model_013_inputs import PrepInput
from abacustest.lib_model.model_012_band import PostBand

from abacustest.constant import RECOMMAND_COMMAND, RECOMMAND_IMAGE, RECOMMAND_MACHINE

from . import (comm_class,
               comm_func,
               comm_report,
               comm_class_exampleSource) 

class StruTypeEnum(String, Enum):
    cif = "cif"
    poscar = "poscar"
    stru = "abacus/stru"

class PPTypeEnum(String, Enum):
    apnsv1 = "apns-V1"
    
def parse_pp_type(pp_type):
    if pp_type == "apns-V1":
        return "/root/pporb/ABACUS-V1/pp","/root/pporb/ABACUS-V1/orb"
    else:
        raise ValueError(f"Unknown pp type: {pp_type}")

class ABACUSImage(String, Enum):
    ltsv310 = "registry.dp.tech/dptech/abacus-stable:LTSv3.10"
    latest_intel = "registry.dp.tech/deepmodeling/abacus-intel:latest"
    latest_gnu = "registry.dp.tech/deepmodeling/abacus-gnu:latest"
    latest_cuda = "registry.dp.tech/deepmodeling/abacus-cuda:latest"

class JobTypeEnum(String, Enum):
    scf = "单点计算"
    relax = "原子弛豫"
    cellrelax = "晶格弛豫"
    band = "能带计算"

def parse_job_type(job_type):
    if job_type == "单点计算":
        return "scf"
    elif job_type == "原子弛豫":
        return "relax"
    elif job_type == "晶格弛豫":
        return "cell-relax"
    elif job_type == "能带计算":
        return "band"
    else:
        raise ValueError(f"Unknown job type: {job_type}")

class BasisTypeEnum(String, Enum):
    pw = "pw"
    lcao = "lcao"

class NewSetting(BaseModel):
    stru_type: StruTypeEnum = Field(default="cif",
                            title="Structure format",
                            description="The format of the structures",)
    
    pp_type: PPTypeEnum = Field(default="apns-V1",
                            title="Pseudopotential and orbital type",
                            description="Please choose one of the pseudopotential and orbital type",)
    
    job_type: JobTypeEnum = Field(default="单点计算",
                            title="Job type",
                            description="Please choose one of the job type",)
    
    basis_type: BasisTypeEnum = Field(default="lcao",
                            title="Basis type",
                            description="Please choose one of the basis type",)
    
    abacus_image: ABACUSImage = Field(default=RECOMMAND_IMAGE,
                          title="Abacus Image",
                          description="The image to run abaucs.",)
    
    abacus_command: String = Field(default=RECOMMAND_COMMAND,
                            title="Abacus Command",
                            description="The command to execute abacus",)
    
    bohrium_machine: String = Field(default=RECOMMAND_MACHINE,
                            title="Bohrium Machine",
                            description="The bohrium machine type to run abacus",)


def post_process(job_type, job_path, work_path):
    """
    Post process the job
    """
    if job_type in ["scf", "relax", "cell-relax"]:
        pass
    elif job_type == "band":
        pwd = os.getcwd()
        os.chdir(work_path)
        PostBand(job_path).run()
        json.dump({
            "band structure of " + i: {"type": "image", "file": os.path.join(i, "band.png")} for i in job_path
        },
        open("supermetrics.json", "w"),indent=4)
        os.chdir(pwd)
    else:
        raise ValueError(f"Unknown job type: {job_type}")

class AutoABACUSModel(
    NewSetting,
    comm_class_exampleSource.ExampleSourceSet,
                    comm_class.OutputSet,
                    comm_class.ConfigSet,
                    BaseModel):
    ...  

def AutoABACUSRunner(opts:AutoABACUSModel) -> int:
    try:
        logs = comm_class.myLog()

        paths = comm_func.create_path(str(opts.IO_output_path))
        output_path = paths["output_path"]
        work_path = paths["work_path"]
        download_path = paths["download_path"]

        logs.iprint("read source setting ...")

        datas = comm_class_exampleSource.read_source(opts,work_path,download_path,logs.iprint)

        pwd = os.getcwd()
        # prepare the fdforce
        os.chdir(work_path)
        allfiles = datas["all_files"]
        
        job_type = parse_job_type(opts.job_type)
        pppath, orbpath = parse_pp_type(opts.pp_type)
        allparams, job_path = PrepInput(allfiles, opts.stru_type, 
                           pp_path=pppath,
                           orb_path=orbpath, 
                           abacus_command=opts.abacus_command,
                           machine=opts.bohrium_machine,
                           image=opts.abacus_image,
                           jobtype=job_type,
                           lcao=True if opts.basis_type == "lcao" else False,
                           ).run()
        
        allparams["config"] = comm_func.read_config(opts)
        allparams["save_path"] = "."
        allparams["bohrium_group_name"] = "AutoAbacus"
        allparams["post_dft"] = {
            "image": "registry.dp.tech/dptech/abacustest:latest",
	        "metrics": {
                "dft_type": "abacus",
                "metrics_name": [
                    {"label": "''.join([i+str({label}.count(i)) for i in set({label})])"},
                    "normal_end",
                    "converge",
                    "nkstot",
                    "ibzk",
                    "nbands",
                    "nelec",
                    "natom",
                    "scf_steps",
                    "total_time",
                    "scf_time",
                    {
                        "scf_time/step": "{scf_time}/{scf_steps}",
                        "ks_solver": "{INPUT}['ks_solver']"
                    },
                    "stress_time",
                    "energy",
                    "energy_per_atom",
                    "force",
                    "stress",
                    "drho_last",
                    "denergy_last",
                    "version"
                ],
                "save_file": "metrics.json"
            }
        }
        
        os.chdir(pwd)

        # execut
        stdout,stderr = comm_func.exec_abacustest(allparams,work_path)
        post_process(job_type, job_path, work_path)
        os.chdir(pwd)
        comm_report.gen_report(opts,logs,work_path,output_path,allparams)
    except:
        traceback.print_exc()
        return 1

    return 0



import numpy as np
def trans_nan_to_none(x):
    if x is None:
        return None
    if isinstance(x, float) and np.isnan(x):
        return None
    return x


def get_bar_option(title, x, y, x_type="category", y_type="value"):
    y_used = [trans_nan_to_none(i) for i in y]
    if y_type == "log":
        # need set 0 to None and set negative to abs
        for ix, v in enumerate(y_used):
            if v in [0, None]:
                y_used[ix] = None
            elif v < 0:
                y_used[ix] = -v

    option = {
        "title": {
            "text": title,
            "textStyle": {
                "color": '#333',
                "fontSize": 18,
                "fontWeight": 'bold'
            }
        },
        "toolbox": {
            "show": True,
            "feature": {
                "dataZoom": {},
                "dataView": {"readOnly": False},
                "magicType": {"type": ['line', 'bar']},
                "restore": {},
                "saveAsImage": {}
            }
        },
        "tooltip": {
            "trigger": 'axis',
            "axisPointer": {
                "type": 'shadow'
            }
        },
        "grid": {
            "left": '3%',
            "right": '4%',
            "bottom": '3%',
            "containLabel": True
        },
        "xAxis": [
            {
                "type": x_type,
                "data": x,
                "axisTick": {
                    "alignWithLabel": True
                },
            }
        ],
        "yAxis": [
            {
                "type": y_type,
                "scale": True
            }
        ],
        "series": [
            {
                "name": 'Direct',
                "type": 'line',
                "barWidth": '60%',
                "data": y_used
            }
        ]
    }
    return option


def produce_multiple_y(title, x, y_list, legend_list, x_type="category", y_type="value"):
    series = []
    y_list = [[trans_nan_to_none(i) for i in y] for y in y_list]
    for iy, y in enumerate(y_list):
        series.append({
            "name": legend_list[iy],
            "type": 'line',
            "barWidth": '60%',
            "data": y.copy(),
            "gridIndex": 0,
            "yAxisIndex": 0,
            "xAxisIndex": 0,
        })
    grid = [{
            "left": '3%',
            "right": '4%',
            "bottom": '3%',
            "top": "25%",
            "containLabel": True
            }]
    xAxis= [
            {
                "type": x_type,
                "data": x,
                "axisTick": {
                    "alignWithLabel": True
                },
                "gridIndex": 0,
            }
        ]
    yAxis = [
            {
                "type": "value",
                "scale": True,
                "gridIndex": 0,
            }
        ]
    if y_type == "log":
        grid = [{
            "left": '3%',
            "right": '55%',
            "bottom": '3%',
            "top": "25%",
            "containLabel": True
        }, {
            "left": '50%',
            "right": '3%',
            "bottom": '3%',
            "top": "25%",
            "containLabel": True
        }]
        xAxis.append({
                "type": x_type,
                "data": x,
                "axisTick": {
                    "alignWithLabel": True
                },
                "gridIndex": 1,
            })
        yAxis.append({
                "type": "log",
                "scale": True,
                "gridIndex": 1,
            })

        # need set 0 to None and set negative to abs
        log_series = []
        for iy, y in enumerate(y_list):
            log_series.append({
                "name": legend_list[iy],
                "type": 'line',
                "barWidth": '60%',
                "data": y.copy(),
                "gridIndex": 1,
                "yAxisIndex": 1,
                "xAxisIndex": 1,
            })
            for ix, v in enumerate(y):
                if v in [0, None]:
                    log_series[iy]["data"][ix] = None
                elif v < 0:
                    log_series[iy]["data"][ix] = -v
        series += log_series

    option = {
        "title": {
            "text": title,
            "textStyle": {
                "color": '#333',
                "fontSize": 18,
                "fontWeight": 'bold'
            }
        },
        "toolbox": {
            "show": True,
            "feature": {
                "dataZoom": {},
                "dataView": {"readOnly": False},
                "magicType": {"type": ['line', 'bar']},
                "restore": {},
                "saveAsImage": {}
            }
        },
        "tooltip": {
            "trigger": 'axis',
            "axisPointer": {
                "type": 'shadow'
            }
        },
        "grid": grid,
        "xAxis": xAxis,
        "yAxis": yAxis,
        "series": series,
        "legend": {"data": legend_list,
                   "top": "10%"}
    }
    return option



import os,sys,glob,time,shutil,argparse,json,traceback,copy,re
from . import globV,comm,PredftOP,PostdftOP,RundftOP
from .. import prepare
from dflow import (
    Workflow,
    Step,
    Steps,
    Inputs,
    Outputs,
    argo_range,
    SlurmRemoteExecutor,
    upload_artifact,
    download_artifact,
    InputArtifact,
    InputParameter,
    OutputArtifact,
    OutputParameter,
    ShellOPTemplate,
    S3Artifact
)

from pathlib import Path
from typing import List
from dflow.plugins.dispatcher import DispatcherExecutor
from dflow.python import upload_packages

from dflow.plugins.bohrium import BohriumContext, BohriumExecutor
from dflow.python import (
    PythonOPTemplate,
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    Slices,
    BigParameter,
    Parameter
)

from dflow import config, s3_config
from dflow.plugins import bohrium
from dflow.plugins.bohrium import TiefblueClient,create_job_group

from abacustest.lib_collectdata.collectdata import RESULT
from abacustest.collectdata import parse_value

def SetConfig(private_set,debug=False):  
    config["archive_mode"] = None
    if debug:
        config["mode"] = "debug"
        host = "LOCAL"
        client = "LOCAL"
    else:
        comm.printinfo("set config info...")
        config["save_keys_in_global_outputs"] = False
        if private_set.get("dflow_host","").strip() != "":
            config["host"] = private_set.get("dflow_host","").strip()
        else:
            config["host"] = "https://workflows.deepmodeling.com"
    
        host = config["host"] + "/workflows/argo"
        
        if private_set.get("dflow_s3_config_endpoint","").strip() != "": 
            s3_config["endpoint"] =  private_set.get("dflow_s3_config_endpoint","").strip()
            comm.printinfo("set s3_config_endpoint ... ")
        
        comm.printinfo("set k8s_api_server ... ")    
        if private_set.get("dflow_k8s_api_server","").strip() != "":
            config["k8s_api_server"] = private_set.get("dflow_k8s_api_server","").strip()
        else:
            config["k8s_api_server"] = "https://workflows.deepmodeling.com"
            
        if private_set.get("dflow_token","").strip() != "":
            config["token"] = private_set.get("dflow_token","").strip()
            comm.printinfo("set token ... ")
            
        if host == "https://lbg-workflow-deepmodeling.dp.tech/workflows/argo":
            config["token"] = ""
            config["namespace"] = "deepmodeling"
            host = "https://lbg-workflow-deepmodeling.dp.tech/workflows/deepmodeling"

        if "bohrium_username" in private_set:
            bohrium.config["username"] = private_set.get('bohrium_username')
            comm.printinfo("set bohrium_username ... ")
            
        if "bohrium_password" in private_set:
            bohrium.config["password"] = private_set.get('bohrium_password','')
            comm.printinfo("set bohrium_password ... ")
        
        if "bohrium_ticket" in private_set:
            bohrium.config["ticket"] = private_set.get('bohrium_ticket')
            comm.printinfo("set bohrium_ticket ... ")
                
        if "bohrium_project_id" in private_set:
            bohrium.config["project_id"] = private_set.get('bohrium_project_id','')
            comm.printinfo("set bohrium_project_id ... ")

        s3_config["repo_key"] = "oss-bohrium"
        s3_config["storage_client"] = TiefblueClient()
        client = s3_config["storage_client"]

        if globV.get_value("BOHRIUM_EXECUTOR"):
            globV.set_value("BRM_CONTEXT",BohriumContext(
                    executor="mixed",
                    extra={},
                    username=bohrium.config["username"],
                    password=bohrium.config["password"]))

        
    globV.set_value("HOST", host)
    globV.set_value("storage_client", client)

def PrepareExample(setting):
    '''
    {
        "ifrun": True,
        "example_template":["example_path"]
        "input_template":"INPUT",
        "kpt_template":"KPT",
        "stru_template":"STRU",
        "mix_input":{
            "ecutwfc":[50,60,70],
            "kspacing":[0.1,0.12,0.13]
        },
        "mix_kpt":[],
        "mix_stru":[],
        "pp_dict":{},
        "orb_dict":{},
        "pp_path": str,
        "orb_path": str,
        "dpks_descriptor":"",
        "extra_files":[],
    }
    '''
    allpath = []
    for ipathlist in prepare.DoPrepare(setting,""):  #put prepared examples to current path
        for ipath in ipathlist:
            allpath.append(ipath)
    return allpath
    
def ParseProcess(param):
    """
    param:{
        "prepare":{},
        "pre_dft":{},
        "run_dft":[{}],
        "post_dft":{}
    }
    """
    prepare_example = pre_dft = run_dft = post_dft = True
    if "prepare" not in param or (not param["prepare"].get("ifrun",True)):
        prepare_example = False
    else:
        prepare_example = param["prepare"]
    
    if "pre_dft" not in param or ( not param["pre_dft"].get("ifrun",True)):
        pre_dft = None
    else:
        pre_dft = param["pre_dft"]
        
    if "run_dft" not in param:
        run_dft = None
    else:
        tmp = []
        run_dft = []
        if isinstance(param["run_dft"],dict):
            tmp = [param["run_dft"]]
        elif isinstance(param["run_dft"],list):
            tmp = param["run_dft"]
        else:
            comm.printinfo("run_dft should be dict or list")
            
        for itmp in tmp:
            if isinstance(itmp,dict):
                if itmp.get("ifrun",True):
                    run_dft.append(itmp)
        if len(run_dft) == 0:
            run_dft = None
        
    if "post_dft" not in param or (not param["post_dft"].get("ifrun",True)):
        post_dft = None
    else:
        post_dft = param["post_dft"]
        
    return prepare_example,pre_dft,run_dft,post_dft 

def ProduceOneSteps(stepname,param):
    prepare_example,pre_dft,run_dft,post_dft = ParseProcess(param)
    save_path = comm.ParseSavePath(param.get("save_path",None))
    steps = Steps(name="abacustesting-steps",
                  outputs=Outputs(artifacts={"outputs" : OutputArtifact()}))
    allstepname, all_save_path = [], []
    has_step = False
    final_step = None
    
    comm.printinfo(f"\n{stepname}")
    comm.printinfo(f"prepare, pre_dft, run_dft, post_dft: {bool(prepare_example)}, {bool(pre_dft)}, {bool(run_dft)}, {bool(post_dft)}")
    
    # if define prepare, then will prepare example firstly
    if prepare_example:
        #return of Doprepare is a list of dict, each dict is an example
        # each element of the dict is a sub example
        example_path = []
        for ii in prepare.DoPrepare(prepare_example,""):
            if ii != None:
                for jj in ii:
                   if prepare.CheckExample(jj,str(ii[jj])):
                       example_path.append(jj)  
        if example_path == []:
            comm.printinfo("WARNING: defined prepare, but no examples matched, skip it!")
            example_path = None 
        elif (not pre_dft) and (not run_dft) and (not post_dft):
            # copy the example to save_path
            for iexample in example_path:
                relative_path = os.path.relpath(iexample,os.getcwd())
                if os.path.isdir(os.path.join(save_path,relative_path)):
                    shutil.rmtree(os.path.join(save_path,relative_path))
                shutil.copytree(iexample,os.path.join(save_path,relative_path))
        #elif post_dft:
        #    # check the path of metrics, if no path, then use example_path as input
        #    if "metrics" in post_dft:
        #        if isinstance(post_dft["metrics"],dict):
        #            if "path" not in post_dft["metrics"]:
        #                post_dft["metrics"]["path"] = example_path
        #        elif isinstance(post_dft["metrics"],list):
        #            for imetric in range(len(post_dft["metrics"])):
        #                if isinstance(post_dft["metrics"][imetric],dict):
        #                    if "path" not in post_dft["metrics"][imetric]:
        #                        post_dft["metrics"][imetric]["path"] = example_path
    else:
        example_path = None  
    
    # in the predft/rundft/postdft step, in not define example, then will use example_path as input
    # if define example, then will use example as input
    predft_step = None
    if pre_dft:
        if not run_dft and not post_dft:
            gather_result = False
        else:
            gather_result = True
            
        step0, allstepname, all_save_path = PredftOP.produce_predft(pre_dft,stepname,example_path,gather_result)
        if step0:
            steps.add(step0)
            has_step = True
            final_step = step0[0]
            predft_step = step0[0]
    
    # post dft need the output of run dft
    # and if has post dft, the output of run dft will be saved in model_output_artifact
    if run_dft:
        if post_dft:
            gather_result = True
        else:
            gather_result = False

        rundft_step,allstepname,all_save_path,output_artifact = RundftOP.produce_rundft(run_dft,predft_step,stepname,example_path,gather_result)
            
        if len(rundft_step) > 0:
            steps.add(rundft_step)
            has_step = True
            final_step = rundft_step[0]
        else:
            comm.printinfo("WARNING: has defined run_dft, but no examples matched, skip it!")
            output_artifact = None
            
    else:
        output_artifact = None
    
    # if output_artifact is not None, it means that the output of run dft will be saved in model_output_artifact
    # and if output_artifact is None, post dft will use the examples defined by example
    if post_dft:
        #postdft_step, allstepname, all_save_path = produce_postdft(post_dft,stepname,output_artifact,save_path)
        postdft_step, allstepname, all_save_path = PostdftOP.produce_postdft(post_dft,output_artifact,stepname,example_path)
        if postdft_step is not None:
            steps.add(postdft_step)
            has_step = True
            final_step = postdft_step[0]

    if final_step:
        steps.outputs.artifacts['outputs']._from = final_step.outputs.artifacts["outputs"]
    
    step = None
    if has_step:
        step = Step(name="abacustesting",template=steps,key="abacustesting") 
    if all_save_path != None:
        all_save_path = [[save_path,i] for i in all_save_path]
    return step,allstepname,all_save_path

def ProduceAllStep(alljobs):
    allstep = []
    allstepname = []
    allsave_path = []

    stepname = alljobs.get("bohrium_group_name","abacustesting")
    step,stepnames,save_path = ProduceOneSteps(stepname,alljobs)
    if step != None:  
        allstep.append(step)
        allstepname.append(stepnames)
        allsave_path.append(save_path) 
        comm.printinfo("\nComplete the preparing for %s.\n" % (stepname))

    return allstep,allstepname,allsave_path






import traceback
import requests,json
import struct
from pprint import pprint
from typing import Iterator
from tqdm  import tqdm
#from loguru import logger
from aim.storage.treeutils import decode_tree

def decode_encoded_tree_stream(stream: Iterator[bytes], concat_chunks=False) -> bytes:
    prev_chunk_tail = b''
    if concat_chunks:
        data = b''
        for chunk in stream:
            data += chunk
        while data:
            (key_size,), data_tail = struct.unpack('I', data[:4]), data[4:]
            key, data_tail = data_tail[:key_size], data_tail[key_size:]
            (value_size,), data_tail = struct.unpack('I', data_tail[:4]), data_tail[4:]
            value, data_tail = data_tail[:value_size], data_tail[value_size:]
            data = data_tail
            yield key, value
    else:
        for chunk in stream:
            data = prev_chunk_tail + chunk
            prev_chunk_tail = b''
            while data:
                try:
                    (key_size,), data_tail = struct.unpack('I', data[:4]), data[4:]
                    key, data_tail = data_tail[:key_size], data_tail[key_size:]
                    (value_size,), data_tail = struct.unpack('I', data_tail[:4]), data_tail[4:]
                    value, data_tail = data_tail[:value_size], data_tail[value_size:]
                    data = data_tail
                except Exception:
                    prev_chunk_tail = data
                    break
                yield key, value
        assert prev_chunk_tail == b''
 
def get_version(run_hash,token):
    params = {
        "record_density": 50,
        "index_density": 5,
    }
    response = requests.post(f'https://tracking.mlops.dp.tech/api/runs/{run_hash}/texts/get-batch',
                        params=params,
                        json=[{'name': "version", 'context': {'datatype': "metrics",'subset': "metrics0"}}],
                        headers={
                            "Content-Type": "application/json",
                            "Authorization": f"Bearer {token}",
                        })
    if response.status_code != 200:
        return None
    try:
        response.raise_for_status()
        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512*1024), concat_chunks=True))
        version = [i[0]["data"] for i in decoded_response["values"]]
        index = decoded_response["iters"]
        return [x[0] for x in sorted(zip(version,index),key=lambda x:x[1])]
    except:
        #print("hash:",run_hash)
        #traceback.print_exc()
        return None
        
def get_experiment_info(experiment_id, token):
    response = requests.get(f'https://tracking.mlops.dp.tech/api/experiments/{experiment_id}/',
                        headers={
                            "Content-Type": "application/json",
                            "Authorization": f"Bearer {token}",
                        })
    response.raise_for_status()
    return response.json()
def get_run_info_batch(experiment, token, offset=None, batch_size=10):
    query = f"run.experiment == '{experiment}' and not run.archived"
    params = {'q': query, 'report_progress': False, 'limit': batch_size, 'exclude_params': True}
    if offset:
        params['offset'] = offset
    response = requests.get('https://tracking.mlops.dp.tech/api/runs/search/run',
                        params=params,
                        headers={
                            "Content-Type": "application/json",
                            "Authorization": f"Bearer {token}",
                        })
    response.raise_for_status()
    decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512*1024), concat_chunks=True))
    return list(sorted(decoded_response.items(), key=lambda i: i[1]['props']['creation_time'], reverse=True))

def collect_from_runinfo(run,token,hash,GetVersion=False):
        metric = {}
        for imetric in run["traces"]["metric"]:
            if imetric["name"].startswith("__"):
                continue
            #only collect super_metrics
            #if imetric["context"].get("subset",None) == "super_metrics0" or \
            #   imetric["context"].get("datatype",None) == "super_metrics":
            metric[imetric["name"]] = [imetric["last_value"]["last"]]
        if GetVersion:
            version = get_version(hash,token)
            if version:
                metric["version"] = version
        return {
            "run_hash": hash,
            "run_name": run["props"]["name"],
            "experiment_name": run["props"]["experiment"]["name"],
            "tags": [i["name"] for i in run["props"]["tags"]],
            "creation_time": run["props"]["creation_time"],
            "end_time": run["props"]["end_time"],
            "metric": metric
            }

def do_collect_info(all_run,all_run_infos, run_infos,token,needed_tags=None,GetVersion=False):
    for hash, run in run_infos:
        if needed_tags:
            tags_in_run = set([i["name"] for i in  run["props"]["tags"]] )
            if not needed_tags.intersection(tags_in_run):
                continue
        all_run_infos.append(collect_from_runinfo(run,token,hash,GetVersion)) #collect selected infos
        all_run.append(run)  #store all infos

def do_collect_metrics(all_run,all_run_infos, run_info, token,needed_tags=None,GetVersion=False):
    if needed_tags:
        tags_in_run = set([i["name"] for i in  run_info[1]["props"]["tags"]] )
        if not needed_tags.intersection(tags_in_run):
            return
    resp = requests.post(f'https://tracking.mlops.dp.tech/api/runs/{run_info[0]}/metric/get-batch',
        data=json.dumps(run_info[1]['traces']['metric']), 
        headers={
            "Content-Type": "application/json",
            "Authorization": f"Bearer {token}",
            })
    run_info_tmp = run_info[1]
    run_info_tmp["metric"] = resp.json()
    all_run.append(run_info_tmp)  #store extra metric values

    selected_run_info = collect_from_runinfo(run_info[1],token,run_info[0],GetVersion)
    for imetric in run_info_tmp["metric"]:
        if imetric["name"].startswith("__"):
            continue
        selected_run_info["metric"][imetric["name"]] = imetric["values"]
    all_run_infos.append(selected_run_info)

def get_runs(token,experiment,experiment_id,needed_tags=None,collect_metrics=False,GetVersion=False):
    import time
    start = time.time()
    batch_size = 45
    offset = None
    all_run_infos = []
    all_run = []
    experiment_info = get_experiment_info(experiment_id, token)
    #logger.info(experiment_info)
    total = experiment_info['run_count']
    needed_tags = set(needed_tags) if needed_tags else None
    with tqdm(total=total) as pbar:
        pbar.set_description(f"fetch next batch offset={offset}, batch_size={batch_size}")
        run_infos = get_run_info_batch(experiment, token, batch_size=batch_size)
        while run_infos:
            if collect_metrics:
                for run_info in run_infos:
                    do_collect_metrics(all_run,all_run_infos, run_info, token,needed_tags,GetVersion)
                    pbar.update(1)
            else:
                do_collect_info(all_run,all_run_infos, run_infos, token,needed_tags,GetVersion)
                pbar.update(batch_size)
            try:
                if len(run_infos) < batch_size:
                    break
                offset = run_infos[-1][0]
                pbar.set_description(f"fetch next batch offset={offset}, batch_size={batch_size}")
                run_infos = get_run_info_batch(experiment, token, offset=offset, batch_size=batch_size)
            except KeyError: # already empty
                break
    elapsed = time.time() - start
    #logger.info(f'All done! elapsed: {elapsed}')
    return all_run,all_run_infos

'''
all_run is a list of all runs (dict), is the original format
one of the run is like:
{
    "props": {
        "name": "lcao-scalapack.lcao-scalapack-86073e.summary",
        "description": null,
        "experiment": {
            "id": "7ab4e46a-43fb-440a-828d-4fbdef5b4709",
            "name": "abacustest/benchmark",
            "description": null
        },
        "tags": [
            {
                "id": "0c23152e-ad9a-4612-aef7-4f94f4a914e4",
                "name": "benchmark-project-abacustest",
                "color": null,
                "description": null
            },
            ...
        ],
        "archived": false,
        "creation_time": 1682010699.0,
        "end_time": 1682010699.462929,
        "active": false
    },
    "traces": {
        "metric": [
            {
                "context": {},
                "name": "__system__cpu",
                "last_value": {
                    "dtype": "float",
                    "first_step": 0,
                    "last_step": 0,
                    "last": 0.0,
                    "version": 2
                }
            },
            ...
        ]
    }
}

The run_infos is a list of the dict of critical informations of each run
one of the dict is like:
{
            "run_name": run["props"]["name"],
            "experiment_name": run["props"]["experiment"]["name"],
            "tags": [i["name"] for i in run["props"]["tags"]],
            "creation_time": run["props"]["creation_time"],
            "end_time": run["props"]["end_time"],
            "metric": metric
}
metric: {metric_nam1:[],
        metric_name2:[],
        ...}

Be noticed, if the collect_metrics is False, the the metric of run_infos is only
the last value. And only set collect_metrics to be True, we can get the all elements of 
each metric.
'''


import numpy as np
from regex import F
from . import comm

def eos_fit(volume,energy):
    # refer to https://github.com/molmod/DeltaCodesDFT
    # fit a common third-order Birch-Murnaghan equation of state to the data
    # volume and energy are lists of volumes and energies
    # return volume, bulk_modulus, bulk_deriv, residuals
    if len(volume) != len(energy):
        print('Error: The number of volumes and energies do not match')
        return None
    v, e = comm.clean_none_list(volume,energy)
    if len(v) < 4:
        print('Error: Not enough data points to fit the equation of state')
        return None
    v,e = zip(*sorted(zip(v,e)))
    
    ve = np.array([v,e]).T
    fitdata = np.polyfit(ve[:, 0] ** (-2. / 3.),ve[:, 1], 3, full=True)
    residuals0 = (fitdata[1] / len(ve[:, 0]))**0.5
    deriv0 = np.poly1d(fitdata[0])
    deriv1 = np.polyder(deriv0, 1)
    deriv2 = np.polyder(deriv1, 1)
    deriv3 = np.polyder(deriv2, 1)
    
    volume0 = 0
    x = 0
    for x in np.roots(deriv1):
        if x > 0 and deriv2(x) > 0:
            volume0 = x**(-3./2.)
            break

    if volume0 == 0:
        print('Error: No minimum could be found')
    
    derivV2 = 4./9. * x**5. * deriv2(x)
    derivV3 = (-20./9. * x**(13./2.) * deriv2(x) -
        8./27. * x**(15./2.) * deriv3(x))
    bulk_modulus0 = derivV2 / x**(3./2.)
    bulk_deriv0 = -1 - x**(-3./2.) * derivV3 / derivV2
    b0 = bulk_modulus0 * 160.21766208 # GPa
    
    fit_x = np.linspace(min(min(ve[:,0]),volume0), max(max(ve[:,0]),volume0), 100)
    fit_y = deriv0(fit_x**(-2./3.))
    e0 = deriv0(volume0**(-2./3.))
    return float(volume0), float(e0), fit_x.tolist(), fit_y.tolist(), float(b0), float(bulk_deriv0), float(residuals0)

def cal_delta(v0, b0, bp, v0_ref, b0_ref, bp_ref):
    # v0, b0, bp: volume, bulk modulus, bulk modulus derivative of the material
    # unit of v0: A^3, unit of b0: GPa, unit of bp: unitless
    
    v01 = v0
    v02 = v0_ref
    b01 = b0 * 10.**9. / 1.602176565e-19 / 10.**30.
    b02 = b0_ref * 10.**9. / 1.602176565e-19 / 10.**30.
    b11 = bp
    b12 = bp_ref
    
    Vi = 0.94 * v01
    Vf = 1.06 * v02
    
    a31 = 9. * v01**3. * b01 / 16. * (b11 - 4.)
    a21 = 9. * v01**(7./3.) * b01 / 16. * (14. - 3. * b11)
    a11 = 9. * v01**(5./3.) * b01 / 16. * (3. * b11 - 16.)
    a01 = 9. * v01 * b01 / 16. * (6. - b11)
    
    a32 = 9. * v02**3. * b02 / 16. * (b12 - 4.)
    a22 = 9. * v02**(7./3.) * b02 / 16. * (14. - 3. * b12)
    a12 = 9. * v02**(5./3.) * b02 / 16. * (3. * b12 - 16.)
    a02 = 9. * v02 * b02 / 16. * (6. - b12)
    
    x = [0, 0, 0, 0, 0, 0, 0]
    
    x[0] = (a01 - a02)**2
    x[1] = 6. * (a11 - a12) * (a01 - a02)
    x[2] = -3. * (2. * (a21 - a22) * (a01 - a02) + (a11 - a12)**2.)
    x[3] = -2. * (a31 - a32) * (a01 - a02) - 2. * (a21 - a22) * (a11 - a12)
    x[4] = -3./5. * (2. * (a31 - a32) * (a11 - a12) + (a21 - a22)**2.)
    x[5] = -6./7. * (a31 - a32) * (a21 - a22)
    x[6] = -1./3. * (a31 - a32)**2.
    
    Fi = np.zeros_like(Vi)
    Ff = np.zeros_like(Vf)
    
    for n in range(7):
        Fi = Fi + x[n] * Vi**(-(2.*n-3.)/3.)
        Ff = Ff + x[n] * Vf**(-(2.*n-3.)/3.)
    
    # delta unit in eV
    delta = np.sqrt((Ff - Fi) / (0.12 * v02))
    return float(delta)

def plot_eos_one(ax,results, results_ref=None, label_size=16, legend_size=14,x_label="Volume ($A^3$)",y_label="Energy ($eV/atom$)"):
    """
    plot the EOS
    ax: the axis to plot
    results and results_ref is a dict:
    {
        "volume": [v1,v2,...],
        "energy_per_atom": [e1,e2,...],
    }
    
    will plot on ax with:
    1. volume vs energy with blue circle
    2. if fit the EOS, plot the EOS-fit with blue dashed line, and the v0 with blue dashed line
    3. if has results_ref: 
        3.1 plot the EOS-ref with red star
        3.2 if fit the EOS-ref, plot the EOS-fit-ref with red dashed line, and the v0_ref with red dashed line
    
    return a dict:
    {
        "fit": None or [v0,e0,fitx,fity,b0,bp,res],
        "fit_ref": None or [v0,e0,fitx,fity,b0,bp,res], # optional
        "delta": None or delta, # optional
    }
    """
    if "volume" not in results or "energy_per_atom" not in results:
        print("No volume or energy_per_atom in results.")
        return {}
    
    x = results["volume"]
    y = results["energy_per_atom"]
    if len(x) != len(y):
        print("The length of volume and energy_per_atom is not equal.")
        return {}
    
    fit1 = eos_fit(x,y)
    if fit1:
        v0, e0, fitx, fity, b0, bp, res = fit1
    else:
        print("Failed to fit the EOS.")
        v0 = fitx = fity = b0 = bp = res = None
        e0 = min(y)
    
    y = np.array(y) - e0
    
    ax.scatter(x,y,label="EOS",marker="o",color="blue")
    if not isinstance(fitx,type(None)) and not isinstance(fity,type(None)) and len(fitx) == len(fity):
        ilabel = "EOS-fit"
        if not isinstance(res,type(None)):
            ilabel += f" (RMSD={res:.2e})"
        ax.plot(fitx,np.array(fity)-e0,label=ilabel,color="blue",linestyle="--")
    
    if not isinstance(v0,type(None)):
        ax.axvline(v0,ls="--",color="blue")
        y_offset = max(y)*0.1
        ax.text(v0,y_offset,f"{v0:.2f}",color="blue",fontsize=legend_size)

    maxy = max(y)

    ax.set_xlabel(x_label,fontsize=label_size)
    ax.set_ylabel(y_label,fontsize=label_size)
    ax.tick_params(axis='x', labelsize=label_size)
    ax.tick_params(axis='y', labelsize=label_size)
    ax.legend(loc="upper right",fontsize=legend_size)
    ax.set_ylim(-maxy*0.1,maxy*1.2)
    ax.grid(True)
    
    delta = None
    ref_fit1 = None
    
    if results_ref:
        ref_x = results_ref.get("volume",None)
        ref_y = results_ref.get("energy_per_atom",None)
        if ref_x == None or ref_y == None or len(ref_x) != len(ref_y):
            print("The length of ref volume and ref energy_per_atom is not equal.")
            return {"fit":fit1,"fit_ref":None,"delta":None}
        else:
            ref_fit1 = eos_fit(ref_x,ref_y)
            if ref_fit1:
                ref_v0, ref_e0, ref_fitx, ref_fity, ref_b0, ref_bp, ref_res = ref_fit1
            else:
                print("Failed to fit the EOS-ref.")
                ref_v0 = ref_fitx = ref_fity = ref_b0 = ref_bp = ref_res = None
                ref_e0 = min(ref_y)

            ref_y = np.array(ref_y) - ref_e0
            maxy = max(maxy,max(ref_y))

            ax.scatter(ref_x,ref_y,label="EOS-ref",marker="*",color="red")
            if not isinstance(ref_fitx,type(None)) and not isinstance(ref_fity,type(None)) and len(ref_fitx) == len(ref_fity):
                ilabel = "EOS-fit-ref"
                if ref_res != None:
                    ilabel += f" (RMSD={ref_res:.2e})"
                ax.plot(ref_fitx,np.array(ref_fity)-ref_e0,label=ilabel,color="red",linestyle="--")

            if not isinstance(ref_v0,type(None)):
                ax.axvline(ref_v0,ls="--",color="red")
                y_offset = maxy*0.1
                ax.text(ref_v0,y_offset,f"{ref_v0:.2f}",color="red",fontsize=legend_size)

            if None not in [v0,b0,bp,ref_v0,ref_b0,ref_bp]:
                delta = cal_delta(v0,b0,bp,ref_v0,ref_b0,ref_bp)            
            
            ax.set_ylim(-maxy*0.1,maxy*1.2)
            ax.legend(loc="upper right",fontsize=legend_size)
            
    return {"fit":fit1,"fit_ref":ref_fit1,"delta":delta}
        
    


import argparse,os
from . import abacustest,collectdata,outresult,prepare,report,remote,model
from importlib.metadata import version
__version__ = version("abacustest")

def parser():
    my_parser = argparse.ArgumentParser(description="abacustest")
    subparser = my_parser.add_subparsers(dest="command")
    
    abacustest.AbacusTestArgs(subparser.add_parser("submit",help="Submit a workflow"))
    abacustest.AbacusTestCheckStatusArgs(subparser.add_parser("status", help="Check the status of the dflow job"))
    abacustest.AbacusTestDownloadArgs(subparser.add_parser("download", help="Download the results of the dflow job"))
    collectdata.CollectDataArgs(subparser.add_parser("collectdata", help="Collect the specified key from the result files"))
    outresult.OutResultArgs(subparser.add_parser("outresult", help="Screen the result files"))
    prepare.PrepareArgs(subparser.add_parser("prepare", help="Prepare the inputs"))
    report.ReportArgs(subparser.add_parser("report", help="Generate a html report for the results"))
    report.ReportArgs(subparser.add_parser("remote", help = "Push the results to the remote server"))
    model.ModelArgs(subparser.add_parser("model", help = "Prepare and post-process the specified model"))
    return my_parser

def print_head():
    print("\n")
    print("--"*30)
    print("+        ABACUSTEST")
    print(f"+ version: {__version__}")
    print(f"+ GITHUB: https://github.com/pxlxingliang/abacus-test/tree/develop")
    print(f"+ BohriumApp: https://bohrium.dp.tech/apps/abacustest")
    print("--"*30+"\n")
    
    
def main():
    print_head()
    my_parser = parser()
    param = my_parser.parse_args()
    if param.command in ['submit','mlops-submit']:
        abacustest.abacustest(param)
    elif param.command == 'status':
        abacustest.checkstatus(param)
    elif param.command == 'download':
        abacustest.downloadflow(param)
    elif param.command == 'collectdata':
        collectdata.collectdata(param)
    elif param.command == 'outresult':
        outresult.outresult(param)
    elif param.command == 'prepare':
        prepare.PrepareInput(param)
    elif param.command == "report":
        report.Report(param)
    elif param.command == "remote":
        remote.Remote(param)
    elif param.command == "model":
        model.RunModel(param)
    else:
        print(my_parser.parse_args(['-h']))
    
if __name__ == '__main__':
    main()    


import os,sys,argparse,json,copy
import traceback
from abacustest.lib_report import table as tb
from abacustest.myflow import globV

HTML_HEAD = """
<html>
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <style>
        body {
            text-align: left;
        }
        
        .head1{
            font-size: 20px;
            font-weight: bold;
            white-space: pre-wrap;
            line-height: 2; 
            margin-top: 1rem;   
        }
        
        .head2{
            font-size: 18px;
            font-weight: bold;
            white-space: pre-wrap;
            line-height: 2;
        }
        
        .head3{
            font-size: 16px;
            font-weight: bold;
            white-space: pre-wrap;
            line-height: 2;
        }

        .tabletitle{
            font-size: 16px;
            font-weight: bold;
            word-wrap: break-word;
            line-height: 2;
        }
        
        .imagetitle{
            font-size: 16px;
            font-weight: bold;
            word-wrap: break-word;
            line-height: 2;
        }

        .doc {
            font-family: Verdana, sans-serif;
            text-align: left;
            display: inline-block;
            font-size: 16px;
            width: 100%;
            word-wrap: break-word;
            white-space: pre-wrap;
            line-height: 1.5;
            margin-bottom: 0.5rem;
        }
        
        #keys table {
          border-collapse: collapse;
          width: 100%;
        }
        #keys td {
          border: none;
          padding: 5px;
          text-align: left;
          line-height: 0.8;
        }
        
        img {
            max-width: 600px;
            max-height: 600px;
            height: auto;
            cursor: zoom-in;
        }

        .overlay {
            display: none;
            position: fixed;
            top: 0;
            left: 0;
            right: 0;
            bottom: 0;
            background-color: rgba(0, 0, 0, 0.8);
            z-index: 1000;
            overflow: auto;
        }

        .overlay img {
            position: absolute;
            top: 50%;
            left: 50%;
            transform: translate(-50%, -50%);
            max-width: 100%;
            max-height: 100%;
            object-fit: contain;
            cursor: zoom-out;
        }
    </style>
</head>
"""

def _table2html(table,has_head=True):    
    # add title
    html = '''\t<table border="2px">\n'''

    # add table head
    start_row = 0
    if has_head:
        html += '\t\t<thead><tr>'
        for i in range(len(table[0])):
            html += '<th>%s</th>' % table[0][i]
        html += '</tr></thead>\n'
        start_row = 1
    
    # add table body
    html += '\t\t<tbody>'
    for i in range(start_row,len(table)):
        html += '\t\t\t<tr>'
        for j in range(len(table[i])):
            html += '<td>%s</td>' % table[i][j]
        html += '</tr>\n'
    html += '\t\t</tbody>\n'
    html += '\t</table>\n'
    
    return html
    

def metrics2html(metrics_set):
    metric_file = metrics_set.get("content","")
    criteria = metrics_set.get("criteria",{})
    title = metrics_set.get("title","")
    metrics = metrics_set.get("metrics",[])
    sort = metrics_set.get("sort",[])
    center = metrics_set.get("center",True)
    
    if not os.path.exists(metric_file):
        print(f"Error: {metric_file} does not exist!")
        return ""
    
    table = tb.file2table(metric_file)
    if not table:
        print(f"Error: transfer {metric_file} to table failed!")
        return ""
    if table in [[],None]:
        return ""
    if not sort: sort = [table[0][0]]
    table, pass_num = tb.format_table(table, metrics, sort, criteria)
    
    html = ""
    if title:
        html += f'''\t<div class="tabletitle">{title}</div>\n'''
        
    if criteria:
        passnum = pass_num["all"]["pass"]
        totalnum = pass_num["all"]["total"]
        icolor = "green" if passnum == totalnum else "red"
        # if all passed, then color the number to green, else red
        html += f'''\t<div class="head2">Pass/Total: <font color="{icolor}">{passnum}/{totalnum} ({passnum/totalnum*100:.2f}%)</font></div>\n'''
        html += tb.gen_criteria(criteria,pass_num)
    html += _table2html(table,has_head=True)
    
    if center:
        html = "\t<center>\n" + html + "\t</center>\n"
    return html

def supermetrics2html(supermetrics_set):
    metric_file = supermetrics_set.get("content","")
    criteria = supermetrics_set.get("criteria",{})
    title = supermetrics_set.get("title","")
    center = supermetrics_set.get("center",True)
    
    if not os.path.exists(metric_file):
        print(f"Error: {metric_file} does not exist!")
        return ""
    
    sm = json.load(open(metric_file))
    if not sm:
        print("Error: read supermetrics failed!")
        return ""
    
    html = ""
    if title:
        html += f'''\t<div class="tabletitle">{title}</div>\n'''
    html += tb.gen_criteria_sm(criteria,sm)
    if center:
        html = "\t<center>\n" + html + "\t</center>\n"
    return html

def table2html(table_set):
    filename = table_set.get("content","")
    title = table_set.get("title","")
    center = table_set.get("center",True)
    
    if not os.path.exists(filename):
        print(f"Error: {filename} does not exist!")
        return ""
    
    html = ""
    if title:
        html += f'''\t<div class="tabletitle\">{title}</div>\n'''
        
    filetype = os.path.splitext(filename)[1]
    if filetype == ".csv":
        table = tb.csv2table(filename)
    else:
        print(f"Error: file type '{filetype}' of table is not supported!")
        return ""

    html += _table2html(table,has_head=True)
    if center:
        html = "\t<center>\n" + html + "\t</center>\n"
    return html

def image2html(image_set):
    image_file = image_set.get("content","")
    title = image_set.get("title","")
    center = image_set.get("center",True)
    
    if isinstance(image_file,str):
        image_file = [image_file]
        
    html = ""

    for ifile in image_file:
        if not os.path.exists(ifile):
            print(f"Error: image '{ifile}' does not exist!")
        html += f"""\t<img class="thumbnail" src="{ifile}" onclick="openFullscreen(this)">\n"""
    
    if title != "":
        html += f'''\t<div class="imagetitle">{title}</div>\n'''
    if center:
        html = "\t<center>\n" + html + "\t</center>\n"
    return html
    
def text2html(text_set):
    '''
    Transform a text to html format
    '''
    text = text_set.get("content","")
    if isinstance(text, str):
        text = text.split("\n")
    elif isinstance(text, list):
        pass
    else:
        print(f"Error: the content of text should be str or list, but {type(text)} is given!")
        return ""
    html = ""
    for ic in text:    
        html += f'''\t<div class="doc">    {ic}</div>\n'''
    return html   

def get_job_address():
    job = globV.get_value("JOB_ADDRESS","")
    if job != "":
        job = f'''<a href="{job}">link</a>'''
    return job

def get_version():
    if os.path.isfile("version.dat"):
        with open("version.dat") as f:
            version = f.read().strip()
    else:
        version = ""
    return version
    
def get_test_date():
    from datetime import datetime
    run_date = globV.get_value("START_TIME",datetime.now()).strftime("%Y-%m-%d")
    return run_date
         
def keys2html(keys):
    html = ""
    html += """\t<table id="keys">\n"""
    comm_keys = ["test_date","version"]
    #if not keys.get("job_address",""):
    #    keys["job_address"] = get_job_address()
        
    if not keys.get("version",""):
        keys["version"] = get_version()
    if not keys.get("test_date",""):
        keys["test_date"] = get_test_date()
        
    # write the comman keys
    for ikey in comm_keys:
        html += f'''\t\t<tr><td><strong>{ikey.capitalize()}</strong></td><td>:</td><td>{keys.get(ikey,"")}</td></tr>\n'''
    # write the other keys
    for ikey in keys.keys():
        if ikey not in comm_keys:
            html += f'''\t\t<tr><td><strong>{ikey.capitalize()}</strong></td><td>:</td><td>{keys.get(ikey,"")}</td></tr>\n'''
    html += """\t</table>\n"""
    return html

def gen_script(has_image=False):
    if not has_image:
        return ""
    
    html = ""
    
    if has_image:
        # add script for image zoom
        html += '''
    <div class="overlay" id="overlay" onclick="closeFullscreen()">
        <img id="fullscreenImage" src="">
    </div>
    '''
    
    html += '''\t<script>\n'''
    
    if has_image:
        html += '''
        var overlay = document.getElementById("overlay");
        var fullscreenImage = document.getElementById("fullscreenImage");
        
        function openFullscreen(imgElement) {
            fullscreenImage.src = imgElement.src; 
            overlay.style.display = "block";
        }
        
        function closeFullscreen() {
            overlay.style.display = "none";
            fullscreenImage.src = ""
        }
        '''
        
    html += '''
    </script>
    '''
    
    return html

def gen_html(report_setting, output):
    '''
    A report should be like this:
    Test Date/Version/Targets/Datasets/Properties/Criteria/Job Address:
    
    Content Section:
    '''
    
    '''
    Keys shold be a dict:
    {
        "test_date": "2020-01-01",
        "version": "1.0",
        "targets": "target1",
        "datasets": "dataset1",
        "properties": "property1",
        "criteria": "criteria1",
        "job_address": "job_address1"
    }
    
    The content section should has the following format:
    head1:
    head2:
    head3:
    text:
    image:
    table:
    metrics:
    supermetrics:
    
    If the content is a table, it can have an extra criteria to color the value, and if the value pass the criteria, the value will be colored to green, else red.
    If the content is a table or image, it can have an extra title.
    If the content is text, it can be a list of str, and each str will be a paragraph.
    
    Should be a list of dict:
    {
        "type": "head1",
        "content": "This is a head1"
    },
    {
        "type": "text",
        "content": "This is a text" or ["This is a text1","This is a text2"]
    }
    {
        "type": "image",
        "content": "image1.png",
        "title": "This is a image"
    },
    {
        "type": "table",
        "title": "This is a table",
        "content": "table1.csv"
    },
    {
        "type": "metrics",  # the row of a metrics table should be the example name, and the column should be the metric name
        "title": "This is a metrics table",
        "content": "table1.csv"
        "criteria":{
            "key1": "x > 0",
            "key2": "x < 1", # key is the name of the column, and the criteria is a string, should be evaluated by python
        },
        "sort": ["key1","key2"] # sort the table based on the key, default is the first column
        "metrics": ["metric1","metric2",] # only show the metrics in the list, default is all
    },
    {
        "type": "supermetrics",  # the row of a metrics table should be the example name, and the column should be the metric name
        "title": "This is a metrics table",
        "content": "table1.csv"
        "criteria":{
            "key1": "x > 0",
            "key2": "x < 1", # key is the name of the column, and the criteria is a string, should be evaluated by python
        }
    }
    '''
    
    keys = report_setting.get("keys",{})
    content = report_setting.get("content",[])
    
    html = HTML_HEAD + "\n<body>\n"
    html += keys2html(keys) + "\n"

    # write content
    has_image = False
    for item in content:
        itype = item.get("type","text")
        icontent = item.get("content","")
        if itype in ["head1","head2","head3"]:
            html += f'''\t<div class="{itype}">{icontent}</div>\n'''
        elif itype == "text":
            html += text2html(item)
        elif itype == "image":
            html += image2html(item)
            has_image = True
        elif itype == "table":
            html += table2html(item)
        elif itype == "metrics":
            html += metrics2html(item)
        elif itype == "supermetrics":
            html += supermetrics2html(item)
        else:
            continue
        html += "\n"
    
    # add script for image zoom
    html += gen_script(has_image)
         
    html += """\n</body>\n</html>"""
    
    with open(output,"w") as f:
        f.write(html)
    return html

def ReportArgs(parser):  
    parser.description = "Read metrics.json and generate the report"
    parser.add_argument('-p', '--param', type=str, help='the parameter file, should be .json type',required=True)
    parser.add_argument('-o', '--output', type=str,  default="abacustest.html",help='The output file name, default is abacustest.html')
    return parser

def Report(param):
    globV._init()
    param_file = param.param
    if not os.path.exists(param_file):
        print(f"Error: {param_file} does not exist!")
        sys.exit(1)
    report_setting = json.load(open(param_file)).get("report",{})

    if report_setting == {}:
        print("Error: report section is empty!")
        sys.exit(1)
    output = param.output
    gen_html(report_setting,output)
    

def main():
    parser = argparse.ArgumentParser()
    param = ReportArgs(parser).parse_args()
    Report(param)
    
if __name__ == "__main__":
    main()


import os
from . import globV,comm
from dflow import (
    Step,
    upload_artifact,
    S3Artifact
)

from pathlib import Path
from typing import List, Optional

from dflow.python import (
    PythonOPTemplate,
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    BigParameter
)

from . import comm

class PreDFT(OP):
    def __init__(self):
        pass

    @classmethod
    def get_input_sign(cls):
        return OPIOSign(
            {
                "examples": Artifact(List[Path],optional=True),
                "command": str,
                "extra_files": Artifact(Path,optional=True),
                "predft_setting" : BigParameter(dict,default={}),  #the setting of predft
            }
        )

    @classmethod
    def get_output_sign(cls):
        return OPIOSign(
            {
                "outputs": Artifact(List[Path]),
                "work_directories": [List[str]]  #a list of a list of work directories in rundft
            }
        )

    @OP.exec_sign_check
    def execute(
        self,
        op_in: OPIO,
    ) -> OPIO:
        
        print(op_in)
        log = ""
        work_directories = []
        outputs = []
        if not op_in["examples"]:
            all_examples = [os.getcwd()]
            example_root = None
        else:
            all_examples = op_in["examples"]
            example_root = op_in["examples"].art_root
        
        for iexample in all_examples:
            if example_root != None:
                #current_path = str(iexample).split("inputs/artifacts/examples/")[1]
                current_path = os.path.relpath(str(iexample),str(example_root))
                root_path = False
            else:
                current_path = ""
                root_path = True
            #work_directories.append([])
            log += "\nPrepare example %s\n" % current_path
            if op_in["extra_files"] != None:
                #extra_file_path = str(op_in["extra_files"]).split("/inputs/artifacts/extra_files")[0] + "/inputs/artifacts/extra_files"
                extra_file_path = op_in["extra_files"].art_root
                comm.CopyFiles(extra_file_path,iexample,move=False)

            os.chdir(iexample)
            log += "COMMAND: %s\n" % str(op_in["command"])
            if op_in["command"] != None and op_in["command"].strip() != "":
                cmd = "ulimit -c 0; " + str(op_in["command"])
                return_code, out, err = comm.run_command(cmd)
                log += out + err
                #log += os.popen("(%s) 2>&1" % cmd).read()

            work_directories_filename = op_in["predft_setting"].get("work_directories_filename") 
            if not work_directories_filename:
                work_directories_filename = "example.txt"
                print(f"Have not define the work directories filename! Try to find the work directories in file '{work_directories_filename}'")            
            
            if os.path.isfile(work_directories_filename):
                with open(work_directories_filename) as f:
                    lines = f.readlines()
                for i in lines:
                    if not os.path.isdir(i.strip()):
                        print("Can not find work directory %s" % i.strip())
                    else:
                        #work_directories[-1].append(os.path.join(current_path,i.strip()))
                        work_directories.append(os.path.join(current_path,os.path.relpath(i.strip(),os.getcwd())))
                        
                        if root_path:
                            outputs.append(Path(os.path.relpath(i.strip(),os.getcwd())))
                        else:
                            outputs.append(Path.resolve(Path(i.strip())))
            else:
                print(f"Can not find work_directories_filename '{work_directories_filename}'! Return current path: {iexample}")
                if root_path:
                    outputs.append(Path(os.path.relpath(iexample,os.getcwd())))
                else:
                    outputs.append(Path.resolve(Path(iexample)))
                #current_path = str(iexample).split("inputs/artifacts/dflow_examples")[1] #.split("/",1)[1]
                #current_path = "." if "/" not in current_path.strip().strip("/") else current_path.split("/",1)[1]
                #work_directories[-1].append(current_path)
                work_directories.append(current_path)
                
        
        print(outputs)
        print(work_directories)
        return OPIO({
            "outputs": outputs,
            "work_directories": work_directories
        })


def produce_predft(predft_set,stepname,example_path,gather_result=False):
    #predft
    #produce step
    allsteps = []
    allstepname = []
    all_save_path = []
    
    comm.printinfo("\nPreparing pre_dft")
    #assert example_path or predft_set.get("example") != None, "example in predft is not defined"
    if "example" not in predft_set:
        example_list = example_path if example_path else []
        example_source = None
        example_source_type = "local"
    else:
        example_list = predft_set["example"]
        example_source = predft_set.get("example_source")
        example_source_type = predft_set.get("example_source_type", "local")
    examples, examples_name = comm.transfer_source_to_artifact(
        example_list,
        source=example_source,
        source_type=example_source_type,
        only_folder=True,
        oneartifact=False)
    
    extrafiles, extrafiles_name = comm.transfer_source_to_artifact(
        predft_set.get("extra_files", []),
        source=predft_set.get("extra_files_source"),
        source_type=predft_set.get("extra_files_source_type", "local"),
        only_folder=False,
        oneartifact=True)
    
    if not examples:
        comm.printinfo("No examples found!")
    
    executor, bohrium_set = comm.ProduceExecutor(predft_set, group_name=stepname + "/predft")
    image = globV.get_value("ABBREVIATION").get(predft_set.get("image"), predft_set.get("image"))
    model_output_artifact = S3Artifact(key="{{workflow.name}}/predft" )
    
    #ngroup = predft_set.get("ngroup",1)
    #if ngroup != None and ngroup < 1:
    #    ngroup = len(examples)
    group_size = predft_set.get("group_size",len(examples_name))
    if group_size != None and group_size < 1:
        group_size = len(examples_name)
    
    igroup = 0
    istep = 0
    if examples:
        new_examples,new_examples_name = comm.SplitGroupSize(examples,examples_name,group_size)
    else:
        new_examples_name = [[]]
    for iexample_name in new_examples_name:
        igroup += 1
        dflow_stepname = f"predft-{igroup}"
        bohri_stepname = stepname+f"/predft-{igroup}"
        space = "\n" + (len(bohri_stepname)+2)*" "
        comm.printinfo("%s: %s" % (bohri_stepname,space.join(iexample_name)))
        pt = PythonOPTemplate(PreDFT,image=image,envs=comm.SetEnvs())
        artifacts={}
        if iexample_name:
            artifacts["examples"] = upload_artifact(iexample_name,archive=globV.get_value("COMPRESS"))

        #artifacts={"examples": upload_artifact(iexample_name,archive=None)}
        
        #get extrafiles
        if extrafiles:
            artifacts["extra_files"]=extrafiles[0][0]

        step = Step(name=dflow_stepname, template=pt,
                    parameters={
                        "command": predft_set.get("command",""),
                        "predft_setting" : {"work_directories_filename":predft_set.get("work_directories_filename")}
                    },
                    artifacts=artifacts)
        if executor != None:
            step.executor = executor
        if gather_result:
            step.outputs.artifacts["outputs"].save = [model_output_artifact]
            step.template.outputs.artifacts["outputs"].archive = None
        allsteps.append(step)
        allstepname.append(dflow_stepname)
        all_save_path.append("")
    
    '''
    If use slice, the examples can only be an artifact of List[Path], and if use group_size, the group size should divide len(examples)
    We try to support user specified parallel measurement, including serial calculation of specified examples on one machine
    '''
    comm.printinfo("image: %s" % image)
    comm.printinfo("set bohrium: %s" % str(bohrium_set))
    comm.printinfo("command: %s" % str(predft_set.get("command")))
    comm.printinfo("work_directories_filename: %s" % str(predft_set.get("work_directories_filename")))

    #allstepname = [stepname]
    #all_save_path = [[save_path, ""]]
    return allsteps, allstepname, all_save_path


